/*

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of Stanford University nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
// SafeHistory
// Ensures that visited link coloration cannot be used to 
// bypass your cookie preferences.
//
// Author: Collin Jackson
// Other contributors: Andrew Bortz, Dan Boneh, John Mitchell
//

//////////////////////////////////////////////////////////////////////////////
// Constants

const kSSH_ENABLED_PREF = "stanford-safehistory.enabled";
const kSSH_COOKIE_BEHAVIOR_PREF = "network.cookie.cookieBehavior";
const kSSH_REFERRER_URL = "http://www.safehistory.com/referrer?";

////////////////////////////////////////////////////////////////////////////
// Debug stuff

/**
 * Dump information to the console?
 */
var SSH_debug = true;

/**
 * Sends data to the console if we're in debug mode
 * @param msg The string containing the message to display
 */
function SSH_dump(msg) {
  if (SSH_debug)
    dump("|||||||||| SSH: " + msg + "\n");
}

////////////////////////////////////////////////////////////////////////////
// "Major" objects/classes

/**
 * Referrer History
 * Wrapper object for RDF referrer datasource
 */

function SSH_ReferrerHistory(controller) {
  this.controller = controller;
}

SSH_ReferrerHistory.prototype =
{
  // Return true if the given referrer-uri combination appears in our
  // database (or another referrer from the same host)
  check: function(uri, referrer) {
    var referrer_mismatch = false;
    var RDF = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                        .getService(Components.interfaces.nsIRDFService);
    var history = RDF.GetDataSource("rdf:history");
    var source = RDF.GetResource(uri.spec);
    var property = 
      RDF.GetResource("http://home.netscape.com/NC-rdf#Referrer");
    var targets = history.GetTargets(source, property, true);
    while (targets.hasMoreElements()) {
      var target = targets.getNext();
      target.QueryInterface(Components.interfaces.nsIRDFResource);
      if(target.Value.indexOf(kSSH_REFERRER_URL == 0)) {
        {
          var oldreferrerlist = target.Value.split("?")[1];
          var oldreferrers = oldreferrerlist.split("&");
          for (var i = 0; i < oldreferrers.length; i++) {
            if (referrer.host == 
                this.controller.makeURI(unescape(oldreferrers[i])).host) {
              return true;
            } else referrer_mismatch = true;
          }
        }
      } else SSH_dump("Missing referrer info for " + uri);
    }
    if (referrer_mismatch) {
      SSH_dump("Coloring visited link " + uri.spec + 
               " from " + referrer.host + 
               " disallowed due to lack of referrer in" +
               " history database.");
    }
    return false;
  },
}

/**
 * Load Listener
 * Watches for pages to load and marks links visited
 */

function SSH_LoadListener(controller) {
  this.controller = controller;
  var contentDocument = document.getElementById('content');
  var self = this;
  var closure = function(evt) { self.maybeColorLinks(evt); };
  contentDocument.addEventListener('load', closure, true);

  this.history = 
    Components.classes["@mozilla.org/browser/global-history;2"]
              .getService(Components.interfaces.nsISafeHistory);
}

SSH_LoadListener.prototype =
{
  controller: null,

  history: null,

  // Decide whether it safe to color links on the current page
  maybeColorLinks: function(evt) {
    if(!this.controller.getEnabled()) 
      return;  // Links will be colored by default
    switch(this.controller.getCookieBehavior()) {
      case this.controller.REJECT_COOKIES:
        return;  // Links should never be colored
      case this.controller.NO_FOREIGN_COOKIES:
        var topHref = 
          evt.originalTarget.defaultView.top.document.location.href;
        var curHref = evt.originalTarget.location.href;
        if(topHref == curHref || 
           this.controller.makeURI(topHref, null).host == 
           this.controller.makeURI(curHref, null).host) {
          break;  // Third party check passed
        } else {
          return;  // Third party check failed
        }
     case this.controller.ALLOW_COOKIES:
        break;  // Proceed with same-origin link coloration
    }
    var target = evt.originalTarget;
    var locationURI = this.controller.makeURI(target.location, null);
    if(target.getElementsByTagName) {
      var links = target.getElementsByTagName('a');
      // SSH_dump("Proceeding with coloration for " + locationURI.spec);
      try {
        this.colorLinks(links, locationURI);
      } catch(e) { SSH_dump(e); }
    }
  },

  // Color an array of links at a particular location
  colorLinks: function(links, locationURI) {
    for(var i = 0; i < links.length; i++) {
      try {
        var node = links[i];
	this.colorLink(node, locationURI);  // Color it now

	// Also set up to color it later when it is visited
        var self = this;
        var closure = function() { self.colorLink(node, locationURI); };
	node.addEventListener("click", closure, false);

      } catch(e) {
        SSH_dump("Error coloring links on " + locationURI.spec + ": " + e);
      }
    }
    // SSH_dump("Done coloring " + i + " links.");
  },

  // Color a single link at a particular location
  colorLink: function(node, locationURI) {
    var oldHref = node.getAttribute("href");
    if(typeof oldHref != "string") return;
    var hrefURI = this.controller.makeURI(oldHref, locationURI);
    if(hrefURI.host == locationURI.host ||
       this.controller.referrerHistory.check(hrefURI, locationURI)) {
      node.setAttribute("href", "");
      this.history.isVisitedLinkAllowed = true;  // Enable real coloration 
      node.setAttribute("href", oldHref);
      this.history.isVisitedLinkAllowed = false; // Disable real coloration
      SSH_dump("Can color " + hrefURI.spec);
    } else {
      SSH_dump("Cannot color offsite link " + hrefURI.spec);
    }
  }
}

/**
 * Master control object. Adds the RequestListener and LoadListener
 */

function SSH_Controller() {
  this.referrerHistory = new SSH_ReferrerHistory(this);
  this.loadListener = new SSH_LoadListener(this);  
  SSH_dump("Loaded controller");
}

SSH_Controller.prototype = {

  loadListener: null,
  referrerHistory: null,

  getEnabled: function() {
    return Components.classes["@mozilla.org/preferences-service;1"]
                     .getService(Components.interfaces.nsIPrefBranch)
                     .getBoolPref(kSSH_ENABLED_PREF);
  },

  // Returns the value of the network.cookie.cookieBehavior pref
  ACCEPT_COOKIES: 0,
  NO_FOREIGN_COOKIES: 1,
  REJECT_COOKIES: 2,
  getCookieBehavior: function() {
    return Components.classes["@mozilla.org/preferences-service;1"]
                     .getService(Components.interfaces.nsIPrefService)
                     .getIntPref(kSSH_COOKIE_BEHAVIOR_PREF);
  },

  // Utility function to construct an nsIURI
  makeURI: function(href, baseURI) {
    var ioService = 
      Components.classes["@mozilla.org/network/io-service;1"]
                .getService(Components.interfaces.nsIIOService);
    try {
      if(baseURI) return ioService.newURI(baseURI.resolve(href),null,null);
      else return ioService.newURI(href, null, null);
    } catch(ex) {
      return null;
    }
  },
}

/////////////////////////////////////////////////////////////////////////
// Global stuff
// "What script would be complete without a couple of globals?" --Fritz
var SSH_controller = null;

function SSH_startup(event) {
  try {
    if(!SSH_controller) SSH_controller = new SSH_Controller();
  } catch(e) { SSH_dump(e); }
}

window.addEventListener("load", SSH_startup, false);
